import numpy as np
import numpy.typing as npt
from tram import mat_linear_assign, mat_linear_congested_assign

from .entities import AllStopService, LimitedStopService
from .instance import LSSDPInstance
from .utils import calculate_stats, trip_time


class LSSDPSolution:
    def __init__(self, inst: LSSDPInstance) -> None:
        self._inst = inst

        self._ass = AllStopService(
            inst.nstops, inst.nbuses - 1, inst.travel_time, inst.capacity
        )
        self._lss = LimitedStopService(inst.nstops, 1, inst.travel_time, inst.capacity)

        self._prev_obj = 1.0
        self._obj = 1.0
        self._ttd = inst.base_ttd
        self._flow = inst.base_flow

    @property
    def _lf(self) -> npt.NDArray[np.floating]:
        return (
            np.divide(self._flow, self._capacities)
            if self._inst.congested
            else np.array([float("nan")], dtype=np.float32)
        )

    @property
    def _ass_load_factor(self) -> npt.NDArray[np.floating]:
        return (
            self._ass.convert_invehicle_flow_to_mat(self._lf)
            if self._inst.congested
            else np.array([float("nan")], dtype=np.float32)
        )

    @property
    def _lss_load_factor(self) -> npt.NDArray[np.floating]:
        return (
            self._lss.convert_invehicle_flow_to_mat(self._lf)
            if self._inst.congested
            else np.array([float("nan")], dtype=np.float32)
        )

    @property
    def _ass_flow_mat(self) -> npt.NDArray[np.floating]:
        return self._ass.convert_invehicle_flow_to_mat(self._flow)

    @property
    def _lss_flow_mat(self) -> npt.NDArray[np.floating]:
        return self._lss.convert_invehicle_flow_to_mat(self._flow)

    @property
    def _capacities(self) -> npt.NDArray[np.floating]:
        if not self._lss.is_valid():
            return np.array([self._inst._oris.max_load] * 3 * (self._inst.nstops - 1))

        return np.array(
            ([self._ass.max_load] * 3 * (self._ass.nstops - 1))
            + ([self._lss.max_load] * 3 * (self._lss.nstops - 1))
        )

    @property
    def stats(
        self,
    ) -> dict[str, tuple[np.floating, np.floating, np.floating, np.floating]]:
        if not self._lss.is_valid():
            per_flow_exp = 0.0
        else:
            per_flow_exp = (
                self._flow[3 * (self._ass.nstops - 1) :].sum() / self._flow.sum()
            )

        return {
            "ttd": calculate_stats(self._ttd),
            "lf": calculate_stats(self._lf),
            "per_flow_exp": calculate_stats([per_flow_exp]),
        }

    def terminate(self) -> None:
        self._prev_obj = self._obj

    def toggle(self, stop: int) -> None:
        self._lss.toggle(stop)
        self._calculate_objective()

    def add_bus(self) -> None:
        self._ass.remove_bus()
        self._lss.add_bus()
        self._calculate_objective()

    def remove_bus(self) -> None:
        self._ass.add_bus()
        self._lss.remove_bus()
        self._calculate_objective()

    def _calculate_objective(self) -> None:
        if not self._lss.is_valid():
            self._prev_obj = self._obj
            self._obj = 1.0
            self._ttd = self._inst.base_ttd

            return

        alignments = [self._ass.stops, self._lss.stops]
        frequencies = [self._ass.frequency, self._lss.frequency]

        if self._inst.congested:
            out = mat_linear_congested_assign(
                alignments,
                frequencies,
                self._inst.travel_time,
                self._inst.demand,
                self._inst.capacity,
                max_iters=self._inst.max_iters,
            )
        else:
            out = mat_linear_assign(
                alignments,
                frequencies,
                self._inst.travel_time,
                self._inst.demand,
            )

        self._prev_obj = self._obj
        self._obj = out[2] / self._inst.base_obj
        self._ttd = np.asarray(out[0], dtype=np.float32)
        self._flow = np.asarray(out[1], dtype=np.float32)



import numpy as np
import numpy.typing as npt

from .entities import AllStopService as AllStopService
from .entities import LimitedStopService as LimitedStopService
from .instance import LSSDPInstance as LSSDPInstance
from .utils import calculate_stats as calculate_stats
from .utils import trip_time as trip_time

class LSSDPSolution:
    _inst: LSSDPInstance
    _ass: AllStopService
    _lss: LimitedStopService
    _obj: float
    _prev_obj: float
    _ttd: npt.NDArray[np.floating]
    _flow: npt.NDArray[np.floating]
    def __init__(self, inst: LSSDPInstance) -> None: ...
    @property
    def stats(
        self,
    ) -> dict[str, tuple[np.floating, np.floating, np.floating, np.floating]]: ...
    def terminate(self) -> None: ...
    def toggle(self, stop: int) -> None: ...
    def add_bus(self) -> None: ...



import abc
from abc import ABC, abstractmethod
from dataclasses import dataclass as dataclass

import numpy.typing as npt

class Service(ABC, metaclass=abc.ABCMeta):
    def __init__(
        self, nstops: int, nbuses: int, travel_time_mat: npt.NDArray, capacity: float
    ) -> None: ...
    @property
    @abc.abstractmethod
    def stops(self) -> list[int]: ...
    @abc.abstractproperty
    def _invehicle_flow_indices(self) -> list[int]: ...
    @property
    def nbuses(self) -> int: ...
    @property
    def trip_time(self) -> float: ...
    @property
    def frequency(self) -> float: ...
    @property
    def max_load(self) -> float: ...
    @property
    def last_stop(self) -> int: ...
    @property
    def nstops(self) -> int: ...
    def is_valid(self) -> bool: ...
    @abstractmethod
    def is_serving(self, stop: int) -> bool: ...
    def convert_invehicle_flow_to_mat(self) -> npt.NDArray: ...

class AllStopService(Service):
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def stops(self) -> list[int]: ...
    def is_serving(self, stop: int) -> bool: ...
    def not_serving_any_stops(self) -> bool: ...
    def remove_bus(self) -> None: ...

class LimitedStopService(Service):
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def stops(self) -> list[int]: ...
    @property
    def stops_binary(self) -> list[bool]: ...
    def is_serving(self, stop: int) -> bool: ...
    def not_serving_any_stops(self) -> bool: ...
    def add_bus(self) -> None: ...
    def toggle(self, stop: int) -> None: ...



from typing import Annotated, Any

from gymnasium import Env, spaces

from .instance import LSSDPInstance as LSSDPInstance
from .network import SGNetwork as SGNetwork
from .solution import LSSDPSolution as LSSDPSolution

ActType = int
ObsType = dict[str, Any]

class RidesEnv(Env):
    metadata: dict[str, Any]
    observation_space: spaces.Space
    action_space: spaces.Space
    render_mode: str | None
    def __init__(
        self,
        nstops: int | list[int] = [30, 45],
        min_headway: Annotated[float, "min"] = 3.0,
        max_headway: Annotated[float, "min"] = 15.0,
        nbuses_full_min: int = 1,
        truncate: bool = True,
        capacity: float | None = 90.0,
        congested: bool = True,
        speed: Annotated[float, "kmh"] = 17.0,
        dwell_time: Annotated[float, "min"] = 0.5,
        allow_retrospect: bool = False,
        demand_npeaks_max: int = 50,
        demand_peak_conc: float = 2.0,
        demnad_peak_size: float = 150.0,
        demand_factor: float = 1.25,
        max_iters: int = 10000,
        render_mode: str | None = None,
    ) -> None: ...
    def step(
        self, action: ActType
    ) -> tuple[ObsType, float, bool, bool, dict[str, Any]]: ...
    def reset(
        self, *, seed: int | None = None, options: dict[str, Any] | None = None
    ) -> tuple[ObsType, dict[str, Any]]: ...
    def render(self) -> None: ...
    @property
    def sol(self) -> LSSDPSolution: ...
    @property
    def inst(self) -> LSSDPInstance: ...



from typing import Annotated

import numpy as np
import numpy.typing as npt

from .entities import AllStopService as AllStopService
from .network import SGNetwork as SGNetwork
from .utils import calculate_stats as calculate_stats
from .utils import trip_time as trip_time

class LSSDPInstance:
    travel_time: npt.NDArray[np.floating]
    demand: npt.NDArray[np.floating]
    nbuses: int
    capacity: float
    congested: bool
    base_ttd: npt.NDArray[np.floating]
    base_flow: npt.NDArray[np.floating]
    base_obj: float
    max_iters: int
    name: str | None
    _oris: AllStopService
    def __init__(
        self,
        travel_time: npt.NDArray[np.floating],
        demand: npt.NDArray[np.floating],
        nbuses: int,
        capacity: float,
        congested: bool,
        base_ttd: npt.NDArray[np.floating] = ...,
        base_flow: npt.NDArray[np.floating] = ...,
        base_obj: float = 0.0,
        max_iters: int = 10000,
        name: str | None = None,
    ) -> None: ...
    def visualise(self) -> None: ...
    def calculate_id(self) -> None: ...
    def print_summary(self) -> None: ...
    @property
    def nstops(self) -> int: ...
    @staticmethod
    def from_network(
        network: SGNetwork,
        nstops: int | list[int],
        min_headway: float,
        max_headway: float,
        speed: Annotated[float, "kmh"],
        dwell_time: Annotated[float, "min"],
        demand_npeaks_max: int,
        demand_peak_conc: float,
        demand_peak_size: float,
        demand_factor: float,
        congested: bool,
        capacity: float,
        truncate: bool,
        max_iters: int,
        rng: np.random.Generator,
    ) -> LSSDPInstance: ...
    def __hash__(self) -> int: ...



import numpy as np
import numpy.typing as npt


def calculate_stats(
    arr: npt.ArrayLike, sum: bool = True
) -> tuple[np.floating, np.floating, np.floating, np.floating]:
    arr = np.array(arr)

    values = (
        arr[np.triu_indices_from(arr, k=1)]
        if arr.ndim == 2 and arr.shape[0] == arr.shape[1]
        else arr
    )

    return (
        np.min(values),
        np.mean(values),
        np.max(values),
        np.sum(values) if sum else float("nan"),
    )


def trip_time(travel_time: npt.NDArray, stops: list[int]) -> float:
    return sum(travel_time[from_][to_] for from_, to_ in zip(stops[:-1], stops[1:]))



from typing import Optional

import numpy as np

def idx_max(lst, skip: int = 0): ...

class SGNetwork:
    path_to_data: str
    def __init__(self, path_to_data: str = ".rides_sgnetwork") -> None: ...
    def generate_gmap_link(self, route): ...
    def generate_trunk_route(
        self,
        min_num_nodes: int,
        max_num_nodes: Optional[int] = None,
        rng: np.random.Generator | None = None,
        min_terminal_distance: float = 5000.0,
        sample_candidate_threshold_distance: float = 5000.0,
        max_path_detour: float = 1.2,
        max_link_detour: float = 1.3,
        min_search_radius: float = 0.0,
        max_search_radius: float = ...,
        search_distance_factor: float = 0.2,
    ): ...
    def sample_test_route(self): ...
    def sample_real_route(
        self,
        min_num_nodes: int,
        max_num_nodes: Optional[int] = None,
        truncate: bool = True,
        rng: np.random.Generator | None = None,
    ): ...



import numpy as np
import numpy.typing as npt

def calculate_stats(arr: npt.ArrayLike, sum: bool = True) -> tuple[np.floating, np.floating, np.floating, np.floating]: ...
def trip_time(travel_time: npt.NDArray, stops: list[int]) -> float: ...



from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass

import numpy.typing as npt
import numpy as np

# @dataclass
# class StopSequence:
#     stops: list[int]

#     def __len__(self) -> int:
#         return self.stops.__len__()

#     def __hash__(self) -> int:
#         return tuple(self.stops).__hash__()

#     def __getitem__(self, index):
#         return self.stops.__getitem__(index)


class Service(ABC):
    def __init__(
        self, nstops: int, nbuses: int, travel_time_mat: npt.NDArray, capacity: float
    ):
        self._nbuses = nbuses
        self._nstops = nstops

        self._travel_time = travel_time_mat
        self._capacity = capacity

        self._last_stop = -1

    @abstractproperty
    def stops(self) -> list[int]:
        raise NotImplementedError()

    @abstractproperty
    def _invehicle_flow_indices(self) -> list[int]:
        raise NotImplementedError()

    @property
    def nbuses(self) -> int:
        return self._nbuses

    @property
    def trip_time(self) -> float:
        return sum(
            self._travel_time[from_][to_]
            for from_, to_ in zip(self.stops[:-1], self.stops[1:])
        )

    @property
    def frequency(self) -> float:
        return self.nbuses / self.trip_time

    @property
    def max_load(self) -> float:
        return self.frequency * self._capacity

    @property
    def last_stop(self) -> int:
        return self._last_stop

    @property
    def nstops(self) -> int:
        return len(self.stops)

    def is_valid(self) -> bool:
        return self.nstops >= 2 and self.nbuses >= 1

    @abstractmethod
    def is_serving(self, stop: int) -> bool:
        raise NotImplementedError()

    def convert_invehicle_flow_to_mat(
        self, flow: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.floating]:
        if not self.is_valid():
            return np.array([[0.0]], dtype=np.float32)

        nstops = self.stops[-1] + 1
        out = np.zeros((nstops, nstops), dtype=np.float32)
        values = flow[self._invehicle_flow_indices]
        indices = np.array(self.stops[:-1]) * nstops + np.array(self.stops[1:])
        out.put(indices, values)

        return out

    def remove_bus(self) -> None:
        self._nbuses -= 1

    def add_bus(self) -> None:
        self._nbuses += 1


class AllStopService(Service):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._stops = list(range(self._nstops))

    @property
    def stops(self) -> list[int]:
        return self._stops

    def is_serving(self, stop: int) -> bool:
        return stop < self._nstops

    def not_serving_any_stops(self) -> bool:
        return False

    @property
    def _invehicle_flow_indices(self) -> list[int]:
        return list(range(1, (self.nstops - 1) * 3, 3))


class LimitedStopService(Service):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._stops_binary = [False for stop in range(self._nstops)]

    @property
    def stops(self) -> list[int]:
        return [i for i, served in enumerate(self._stops_binary) if served]

    @property
    def stops_binary(self) -> list[bool]:
        return self._stops_binary

    def is_serving(self, stop: int) -> bool:
        return self._stops_binary[stop]

    def not_serving_any_stops(self) -> bool:
        return sum(self._stops_binary) == 0

    def toggle(self, stop: int) -> None:
        self._stops_binary[stop] = False if self._stops_binary[stop] else True

        if stop >= self._last_stop:
            if self._stops_binary[stop]:
                self._last_stop = stop
            elif sum(self._stops_binary) == 0:
                self._last_stop = -1
        else:
            for i in range(self._last_stop, -1, -1):
                if self._stops_binary[i]:
                    self._last_stop = i
                    break

    @property
    def _invehicle_flow_indices(self) -> list[int]:
        return [self._nstops + i for i in range(1, (self.nstops - 1) * 3, 3)]
